Package: apifetch
Type: Package
Title: Token-Authenticated REST API Retrieval Toolkit
Version: 0.1.0
Date: 2026-06-26
Authors@R: c(
    person("André", "Leite", email = "leite@castlab.org", role = c("aut", "cre")),
    person("Hugo", "Vasconcelos", email = "hugo.vasconcelos@ufpe.br", role = "aut"),
    person("Diogo", "Bezerra", email = "diogo.bezerra@ufpe.br", role = "aut"),
    person("Marcos", "Wasilew", email = "marcos.wasilew@gmail.com", role = "aut"),
    person("Carlos", "Amorin", email = "carlos.agaf@ufpe.br", role = "aut"))
Description: A small, dependency-light toolkit for talking to token-authenticated
    REST APIs. It manages authentication tokens in process environment variables
    (never written to disk), builds requests with configurable authentication and
    pagination strategies, and retrieves paginated data either one page at a time
    or in chunks combined into a single tibble. The design is API-agnostic: a
    single 'apifetch_api' profile describes an endpoint together with how it
    authenticates and paginates, so the same verbs work across different services.
License: MIT + file LICENSE
Encoding: UTF-8
Depends: R (>= 4.1.0)
Imports: cli, dplyr, httr2, stats, tibble, utils
Suggests: knitr, rmarkdown, testthat (>= 3.0.0)
Config/testthat/edition: 3
VignetteBuilder: knitr
URL: https://github.com/StrategicProjects/apifetch,
        https://strategicprojects.github.io/apifetch/
BugReports: https://github.com/StrategicProjects/apifetch/issues
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-06-26 19:37:47 UTC; leite
Author: André Leite [aut, cre],
  Hugo Vasconcelos [aut],
  Diogo Bezerra [aut],
  Marcos Wasilew [aut],
  Carlos Amorin [aut]
Maintainer: André Leite <leite@castlab.org>
Repository: CRAN
Date/Publication: 2026-07-02 18:40:19 UTC
Built: R 4.6.0; ; 2026-07-02 19:47:55 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(print,apifetch_api)
export(af_api)
export(af_auth_bearer)
export(af_auth_header)
export(af_auth_query)
export(af_auth_raw)
export(af_fetch)
export(af_fetch_all)
export(af_get_token)
export(af_list_tokens)
export(af_paginate_none)
export(af_paginate_offset)
export(af_remove_token)
export(af_store_token)
export(parse_queries)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# apifetch <img src="man/figures/logo.png" align="right" height="139" alt="apifetch logo" />

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version/apifetch)](https://CRAN.R-project.org/package=apifetch)
[![R-CMD-check](https://github.com/StrategicProjects/apifetch/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/StrategicProjects/apifetch/actions/workflows/R-CMD-check.yaml)
[![Lifecycle:
experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html#experimental)
[![CRAN
downloads](https://cranlogs.r-pkg.org/badges/grand-total/apifetch)](https://CRAN.R-project.org/package=apifetch)
[![Python
sibling](https://img.shields.io/pypi/v/apifetch?label=python%20sibling&logo=python&logoColor=white)](https://pypi.org/project/apifetch/)
<!-- badges: end -->

`apifetch` is a small, dependency-light toolkit for talking to
token-authenticated REST APIs from R. It handles three recurring chores:

1.  **Token management** — store/get/remove/list tokens in process
    environment variables (never written to disk), namespaced per
    service.
2.  **Request building** — pluggable **authentication** and
    **pagination** strategies, bundled into a reusable API *profile*.
3.  **Data retrieval** — fetch one page, or fetch everything in chunks
    row-bound into a single tibble.

It is the generic engine extracted from the
[BigDataPE](https://github.com/StrategicProjects/BigDataPE) package;
BigDataPE is now just one *use case* (see `vignette("bigdatape")`). A
Python sibling lives at
[apifetch-py](https://github.com/StrategicProjects/apifetch-py).

## Installation

``` r
# install.packages("pak")
pak::pak("StrategicProjects/apifetch")
```

## Usage

``` r
library(apifetch)

# 1. Describe the API once: where, how to authenticate, how to paginate.
api <- af_api(
  endpoint   = "https://api.example.com/v1/search",
  service    = "Example",
  auth       = af_auth_bearer(),          # "Authorization: Bearer <token>"
  pagination = af_paginate_offset("query")
)

# 2. Store a token (kept only in this session's environment).
af_store_token("reports", "my-secret-token", service = "Example")

# 3. Fetch.
one_page <- af_fetch(api, "reports", limit = 50)
everything <- af_fetch_all(api, "reports", chunk_size = 1000)
```

### Strategies

Authentication: `af_auth_bearer()`, `af_auth_raw()`, `af_auth_header()`,
`af_auth_query()`.

Pagination: `af_paginate_offset(where = "header" | "query")`,
`af_paginate_none()`. \`\`\`

# Internal utilities -------------------------------------------------------

# Safe coercion to integer.
# Tries to convert a scalar value to integer. If conversion is not possible
# (e.g. character, logical, NA, or a numeric with fractional part) an
# informative error is raised via cli.
# Values that are already integer or Inf pass through unchanged.
.safe_as_integer <- function(x, arg_name) {
  if (is.infinite(x)) return(x)
  if (is.integer(x)) return(x)
  if (is.numeric(x) && length(x) == 1L && !is.na(x) && x == trunc(x)) {
    return(as.integer(x))
  }
  cli::cli_abort(
    "{.arg {arg_name}} must be a whole number or {.val Inf}, not {.cls {class(x)}} ({.val {x}})."
  )
}

# Sanitize a name into the suffix used for environment-variable token storage.
# Transliterates to ASCII (dropping accents) and turns spaces into underscores,
# matching the contract shared by all token functions.
.sanitize_name <- function(x) {
  x_ascii <- iconv(x, from = "", to = "ASCII//TRANSLIT")
  gsub(" ", "_", x_ascii)
}

# Build the environment-variable name for a token: "<service>_<name>".
.token_var <- function(name, service) {
  paste0(.sanitize_name(service), "_", .sanitize_name(name))
}

#' Build a URL with query parameters
#'
#' Appends a named list of query parameters to a base URL, URL-encoding both
#' names and values and dropping parameters whose value is the empty string.
#'
#' @param url The base URL.
#' @param query_list A named list of query parameters.
#' @return The URL with the query string appended (or the base URL unchanged
#'   when there are no parameters to add).
#' @examples
#' parse_queries("https://example.com", list(a = "1", b = "2"))
#' @export
parse_queries <- function(url, query_list) {
  if (length(query_list) == 0) {
    return(url)
  }

  query_list <- query_list[query_list != ""]

  if (length(query_list) == 0) {
    return(url)
  }

  query_string <- paste(
    vapply(names(query_list), function(name) {
      paste0(
        utils::URLencode(name, reserved = TRUE), "=",
        utils::URLencode(as.character(query_list[[name]]), reserved = TRUE)
      )
    }, character(1)),
    collapse = "&"
  )

  paste0(url, "?", query_string)
}

# API profiles: authentication + pagination strategies --------------------
# An `apifetch_api` object describes *where* to call, *how* to authenticate,
# and *how* to paginate. Auth and pagination are pluggable strategy objects so
# the same fetch verbs work against APIs with different conventions.

# ---- Authentication strategies -------------------------------------------

#' Authentication strategies
#'
#' Constructors describing how a token is attached to a request. Pass the result
#' to the `auth` argument of [af_api()].
#'
#' - `af_auth_raw()`: send the token verbatim in a header (default
#'   `Authorization`). This is what the Big Data PE API expects.
#' - `af_auth_bearer()`: send `"Bearer <token>"` in the `Authorization` header.
#' - `af_auth_header()`: send the token in an arbitrary header (e.g.
#'   `X-API-Key`).
#' - `af_auth_query()`: send the token as a URL query parameter.
#'
#' @param header Header name to use.
#' @param prefix String prepended to the token (bearer scheme).
#' @param param Query-parameter name (query scheme).
#' @return An `apifetch_auth` object.
#' @name af_auth
#' @examples
#' af_auth_raw()
#' af_auth_bearer()
#' af_auth_header("X-API-Key")
#' af_auth_query("api_key")
NULL

.auth <- function(apply) structure(list(apply = apply), class = "apifetch_auth")

#' @rdname af_auth
#' @export
af_auth_raw <- function(header = "Authorization") {
  .auth(function(req, token) {
    h <- stats::setNames(list(token), header)
    do.call(httr2::req_headers, c(list(req), h))
  })
}

#' @rdname af_auth
#' @export
af_auth_bearer <- function(header = "Authorization", prefix = "Bearer ") {
  .auth(function(req, token) {
    h <- stats::setNames(list(paste0(prefix, token)), header)
    do.call(httr2::req_headers, c(list(req), h))
  })
}

#' @rdname af_auth
#' @export
af_auth_header <- function(header = "X-API-Key") {
  .auth(function(req, token) {
    h <- stats::setNames(list(token), header)
    do.call(httr2::req_headers, c(list(req), h))
  })
}

#' @rdname af_auth
#' @export
af_auth_query <- function(param = "api_key") {
  .auth(function(req, token) {
    q <- stats::setNames(list(token), param)
    do.call(httr2::req_url_query, c(list(req), q))
  })
}

# ---- Pagination strategies -----------------------------------------------

#' Pagination strategies
#'
#' Constructors describing how `limit`/`offset` are sent with a request. Pass
#' the result to the `pagination` argument of [af_api()].
#'
#' - `af_paginate_offset()`: send `limit`/`offset` either as HTTP headers
#'   (default, as the Big Data PE API expects) or as URL query parameters.
#' - `af_paginate_none()`: send no pagination parameters.
#'
#' Non-positive or infinite values are omitted from the request.
#'
#' @param where Either `"header"` or `"query"`.
#' @param limit_param,offset_param Parameter names to use.
#' @return An `apifetch_pagination` object.
#' @name af_paginate
#' @examples
#' af_paginate_offset("header")
#' af_paginate_offset("query", limit_param = "per_page")
#' af_paginate_none()
NULL

.pagination <- function(apply) {
  structure(list(apply = apply), class = "apifetch_pagination")
}

#' @rdname af_paginate
#' @export
af_paginate_offset <- function(where = c("header", "query"),
                               limit_param = "limit",
                               offset_param = "offset") {
  where <- match.arg(where)
  .pagination(function(req, limit, offset) {
    vals <- list()
    if (!(is.infinite(limit) || limit <= 0)) vals[[limit_param]] <- limit
    if (!(is.infinite(offset) || offset <= 0)) vals[[offset_param]] <- offset
    if (length(vals) == 0) return(req)
    fn <- if (where == "header") httr2::req_headers else httr2::req_url_query
    do.call(fn, c(list(req), vals))
  })
}

#' @rdname af_paginate
#' @export
af_paginate_none <- function() {
  .pagination(function(req, limit, offset) req)
}

# ---- API profile ----------------------------------------------------------

#' Describe an API endpoint
#'
#' Bundles an endpoint URL with its authentication and pagination strategies and
#' a namespace `service` (used to look up tokens). The resulting object is passed
#' to [af_fetch()] and [af_fetch_all()].
#'
#' @param endpoint The base API URL.
#' @param service Namespace used to look up the token (see [af_get_token()]).
#' @param auth An `apifetch_auth` strategy (see [af_auth]). Default
#'   [af_auth_bearer()].
#' @param pagination An `apifetch_pagination` strategy (see [af_paginate]).
#'   Default [af_paginate_offset()].
#' @param drop_cols Character vector of response columns to drop after parsing
#'   (e.g. a status column). Default none.
#' @param connect_hint Optional extra line shown when a connection error occurs
#'   (e.g. a VPN requirement).
#' @return An `apifetch_api` object.
#' @examples
#' af_api(
#'   endpoint = "https://www.bigdata.pe.gov.br/api/buscar",
#'   service = "BigDataPE",
#'   auth = af_auth_raw(),
#'   pagination = af_paginate_offset("header"),
#'   drop_cols = "Mensagem",
#'   connect_hint = "Ensure you are on the PE Conectado network or VPN."
#' )
#' @export
af_api <- function(endpoint,
                   service = "apifetch",
                   auth = af_auth_bearer(),
                   pagination = af_paginate_offset(),
                   drop_cols = character(0),
                   connect_hint = NULL) {
  if (!is.character(endpoint) || !nzchar(endpoint)) {
    cli::cli_abort("{.arg endpoint} must be a non-empty string.")
  }
  if (!inherits(auth, "apifetch_auth")) {
    cli::cli_abort("{.arg auth} must be an {.cls apifetch_auth} object (see {.fn af_auth_bearer}).")
  }
  if (!inherits(pagination, "apifetch_pagination")) {
    cli::cli_abort("{.arg pagination} must be an {.cls apifetch_pagination} object (see {.fn af_paginate_offset}).")
  }

  structure(
    list(
      endpoint = endpoint,
      service = service,
      auth = auth,
      pagination = pagination,
      drop_cols = drop_cols,
      connect_hint = connect_hint
    ),
    class = "apifetch_api"
  )
}

#' @export
print.apifetch_api <- function(x, ...) {
  cli::cli_h3("<apifetch_api>")
  cli::cli_li("endpoint: {.url {x$endpoint}}")
  cli::cli_li("service: {.val {x$service}}")
  if (length(x$drop_cols)) cli::cli_li("drop_cols: {.val {x$drop_cols}}")
  invisible(x)
}

test_that("parse_queries encodes and drops empties", {
  expect_equal(parse_queries("https://x.test", list()), "https://x.test")
  expect_equal(
    parse_queries("https://x.test", list(a = "1", b = "")),
    "https://x.test?a=1"
  )
  expect_match(
    parse_queries("https://x.test", list(`a b` = "c d")),
    "a%20b=c%20d", fixed = TRUE
  )
})

test_that(".safe_as_integer accepts whole doubles and Inf, rejects fractions", {
  expect_identical(apifetch:::.safe_as_integer(50, "x"), 50L)
  expect_identical(apifetch:::.safe_as_integer(Inf, "x"), Inf)
  expect_error(apifetch:::.safe_as_integer(1.5, "x"))
  expect_error(apifetch:::.safe_as_integer("a", "x"))
})

test_that("af_api validates its strategy arguments", {
  expect_error(af_api("https://x.test", auth = "nope"))
  expect_error(af_api("https://x.test", pagination = "nope"))
  expect_error(af_api(""))
})

test_that("af_api stores the configured pieces", {
  api <- af_api(
    "https://x.test",
    service = "S",
    auth = af_auth_raw(),
    pagination = af_paginate_offset("header"),
    drop_cols = "Mensagem"
  )
  expect_s3_class(api, "apifetch_api")
  expect_equal(api$service, "S")
  expect_equal(api$drop_cols, "Mensagem")
})

test_that("pagination strategies attach params correctly", {
  req <- httr2::request("https://x.test")

  hdr <- af_paginate_offset("header")$apply(req, 10L, 5L)
  expect_equal(hdr$headers$limit, 10L)
  expect_equal(hdr$headers$offset, 5L)

  qry <- af_paginate_offset("query")$apply(req, 10L, 0L)
  expect_match(qry$url, "limit=10", fixed = TRUE)

  # Inf / non-positive values are omitted
  none <- af_paginate_offset("header")$apply(req, Inf, 0L)
  expect_null(none$headers$limit)
})

test_that("auth strategies attach the token", {
  # httr2 redacts the default `Authorization` header, so verify the raw/bearer
  # behaviour through non-redacted custom headers.
  req <- httr2::request("https://x.test")
  expect_equal(af_auth_raw("X-Token")$apply(req, "tok")$headers$`X-Token`, "tok")
  expect_equal(af_auth_bearer("X-Token")$apply(req, "tok")$headers$`X-Token`, "Bearer tok")
  expect_equal(af_auth_header("X-API-Key")$apply(req, "tok")$headers$`X-API-Key`, "tok")
  expect_equal(af_auth_query("api_key")$apply(req, "tok")$url, "https://x.test/?api_key=tok")
})

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(collapse = TRUE, comment = "#>", eval = FALSE)

## -----------------------------------------------------------------------------
# library(apifetch)
# 
# bigdatape <- af_api(
#   endpoint     = "https://www.bigdata.pe.gov.br/api/buscar",
#   service      = "BigDataPE",
#   auth         = af_auth_raw(),
#   pagination   = af_paginate_offset("header"),
#   drop_cols    = "Mensagem",
#   connect_hint = "Ensure you are on the PE Conectado network or VPN."
# )

## -----------------------------------------------------------------------------
# af_store_token("dengue", "your-token-here", service = "BigDataPE")
# 
# # A single page of 50 records
# dengue <- af_fetch(bigdatape, "dengue", limit = 50)
# 
# # Everything, in chunks, with a progress message per chunk
# dengue_all <- af_fetch_all(
#   bigdatape, "dengue",
#   chunk_size = 50000,
#   verbosity = 1
# )

# Token management via environment variables -------------------------------
# Tokens are never written to disk; they live only in process environment
# variables named "<service>_<name>". The `service` acts as a namespace so a
# single R session can hold tokens for several different APIs without clashing.

#' Store an API token in an environment variable
#'
#' Stores an authentication token in a process environment variable named
#' `"<service>_<name>"`. The token is never written to disk. If a non-empty
#' variable with that name already exists, the function refuses to overwrite it.
#'
#' @param name The identifier for this token (e.g. a dataset or resource name).
#' @param token The authentication token (character).
#' @param service A namespace prefix grouping tokens for one API. Default
#'   `"apifetch"`.
#' @return Invisibly `NULL`; called for its side effect.
#' @examples
#' bdpe <- af_store_token("dengue", "your-token-here", service = "BigDataPE")
#' @export
af_store_token <- function(name, token, service = "apifetch") {
  if (!is.character(name) || !nzchar(name)) {
    cli::cli_abort("{.arg name} must be a non-empty string.")
  }
  if (!is.character(token) || !nzchar(token)) {
    cli::cli_abort("{.arg token} must be a non-empty string.")
  }

  env_var_name <- .token_var(name, service)

  if (nzchar(Sys.getenv(env_var_name, unset = ""))) {
    cli::cli_alert_warning(
      "The environment variable {.envvar {env_var_name}} is already defined. Not overwriting to avoid data loss."
    )
    return(invisible())
  }

  env_list <- list(token)
  names(env_list) <- env_var_name
  do.call(Sys.setenv, env_list)

  cli::cli_alert_success("Token stored in environment variable: {.envvar {env_var_name}}")
  invisible()
}

#' Retrieve a stored API token
#'
#' Retrieves the token stored for `name` under `service`. Returns `NULL`
#' (with a warning) rather than erroring when no token is found.
#'
#' @inheritParams af_store_token
#' @return The token string, or `NULL` if not found.
#' @examples
#' token <- af_get_token("dengue", service = "BigDataPE")
#' @export
af_get_token <- function(name, service = "apifetch") {
  if (!is.character(name) || !nzchar(name)) {
    cli::cli_abort("{.arg name} must be a non-empty string.")
  }

  env_var_name <- .token_var(name, service)
  token <- Sys.getenv(env_var_name, unset = "")

  if (token == "") {
    cli::cli_alert_warning("No token found for {.val {name}} (service {.val {service}}).")
    return(NULL)
  }

  token
}

#' Remove a stored API token
#'
#' Removes the token stored for `name` under `service`. Does nothing (beyond a
#' warning) when no token is found.
#'
#' @inheritParams af_store_token
#' @return Invisibly `NULL`; called for its side effect.
#' @examples
#' af_remove_token("dengue", service = "BigDataPE")
#' @export
af_remove_token <- function(name, service = "apifetch") {
  if (!is.character(name) || !nzchar(name)) {
    cli::cli_abort("{.arg name} must be a non-empty string.")
  }

  env_var_name <- .token_var(name, service)

  if (nzchar(Sys.getenv(env_var_name, unset = ""))) {
    Sys.unsetenv(env_var_name)
    cli::cli_alert_success("Token removed for {.val {name}} (service {.val {service}}).")
  } else {
    cli::cli_alert_warning("No token found for {.val {name}} (service {.val {service}}).")
  }
  invisible()
}

#' List stored API tokens
#'
#' Returns the names (without the `service` prefix) of all tokens stored for a
#' given `service` in environment variables.
#'
#' @param service A namespace prefix grouping tokens for one API. Default
#'   `"apifetch"`.
#' @return A character vector of token names, empty if none are found.
#' @examples
#' af_list_tokens(service = "BigDataPE")
#' @export
af_list_tokens <- function(service = "apifetch") {
  prefix <- paste0("^", .sanitize_name(service), "_")
  all_envs <- Sys.getenv()
  stored <- grep(prefix, names(all_envs), value = TRUE)

  if (length(stored) == 0) {
    cli::cli_alert_info("No tokens found for service {.val {service}}.")
    return(character(0))
  }

  sub(prefix, "", stored)
}

test_that("store / get / list / remove round-trip works", {
  svc <- "apifetchTest"
  on.exit(suppressMessages(af_remove_token("alpha", service = svc)), add = TRUE)

  suppressMessages(af_store_token("alpha", "tok-123", service = svc))
  expect_equal(af_get_token("alpha", service = svc), "tok-123")
  expect_true("alpha" %in% af_list_tokens(service = svc))

  suppressMessages(af_remove_token("alpha", service = svc))
  expect_null(suppressMessages(af_get_token("alpha", service = svc)))
})

test_that("store refuses to overwrite an existing token", {
  svc <- "apifetchTest"
  on.exit(suppressMessages(af_remove_token("beta", service = svc)), add = TRUE)

  suppressMessages(af_store_token("beta", "first", service = svc))
  suppressMessages(af_store_token("beta", "second", service = svc)) # no overwrite
  expect_equal(af_get_token("beta", service = svc), "first")
})

test_that("names with accents and spaces are sanitized consistently", {
  svc <- "apifetchTest"
  on.exit(suppressMessages(af_remove_token("São Paulo", service = svc)), add = TRUE)

  suppressMessages(af_store_token("São Paulo", "tok", service = svc))
  expect_equal(af_get_token("São Paulo", service = svc), "tok")
})

test_that("get / remove inform (not error) when token is missing", {
  expect_message(res <- af_get_token("missing", service = "nope"))
  expect_null(res)
  expect_message(af_remove_token("missing", service = "nope"))
})

test_that("invalid inputs error", {
  expect_error(af_store_token("", "tok"))
  expect_error(af_store_token("x", ""))
  expect_error(af_get_token(""))
})

# Data fetching -------------------------------------------------------------

#' Fetch a single page from an API
#'
#' Performs one authenticated request against an [af_api()] profile, applying its
#' pagination strategy, and returns the parsed body as a tibble. HTTP errors and
#' connection failures are translated into friendly `cli` messages.
#'
#' @param api An `apifetch_api` object (see [af_api()]).
#' @param name The token name to authenticate with (looked up via the API's
#'   `service`).
#' @param limit Maximum number of records to request. Default `Inf` (no limit).
#'   Non-positive or infinite values omit the parameter.
#' @param offset Starting record. Default `0` (omitted).
#' @param query A named list of additional query-string filters. Default empty.
#' @param verbosity `0` (silent, default), `1` (progress messages), or `2`
#'   (progress plus full HTTP request/response details).
#' @return A tibble with the parsed response.
#' @examples
#' \dontrun{
#' api <- af_api("https://www.bigdata.pe.gov.br/api/buscar",
#'               service = "BigDataPE", auth = af_auth_raw(),
#'               pagination = af_paginate_offset("header"))
#' af_store_token("dengue", "token", service = "BigDataPE")
#' af_fetch(api, "dengue", limit = 50)
#' }
#' @export
af_fetch <- function(api, name, limit = Inf, offset = 0L, query = list(),
                     verbosity = 0L) {
  if (!inherits(api, "apifetch_api")) {
    cli::cli_abort("{.arg api} must be an {.cls apifetch_api} object (see {.fn af_api}).")
  }
  if (!is.list(query)) {
    cli::cli_abort("{.arg query} must be a {.cls list}.")
  }

  token <- af_get_token(name, service = api$service)
  if (is.null(token)) {
    cli::cli_abort("No token available for {.val {name}}; store one with {.fn af_store_token}.")
  }

  offset    <- .safe_as_integer(offset, "offset")
  limit     <- .safe_as_integer(limit, "limit")
  verbosity <- .safe_as_integer(verbosity, "verbosity")

  req <- api$endpoint |>
    parse_queries(query_list = query) |>
    httr2::request() |>
    httr2::req_error(is_error = function(resp) FALSE)

  req <- api$auth$apply(req, token)
  req <- api$pagination$apply(req, limit, offset)

  httr2_verbosity <- if (verbosity >= 2L) 1L else 0L

  resp <- tryCatch(
    httr2::req_perform(req, verbosity = httr2_verbosity),
    error = function(e) {
      msg <- c(
        "Unable to connect to the API at {.url {api$endpoint}}.",
        "i" = "Check your network connection."
      )
      if (!is.null(api$connect_hint)) msg <- c(msg, "i" = api$connect_hint)
      msg <- c(msg, "x" = conditionMessage(e))
      cli::cli_abort(msg)
    }
  )

  status <- httr2::resp_status(resp)
  if (status >= 400L) {
    reason <- httr2::resp_status_desc(resp)
    cli::cli_abort(c(
      "The API returned an error (HTTP {status} - {reason}).",
      "i" = "Try again later, and check that the endpoint is correct and your token is valid."
    ))
  }

  resp |>
    httr2::resp_body_json(simplifyVector = TRUE) |>
    tibble::as_tibble()
}

#' Fetch all data from an API in chunks
#'
#' Iteratively calls [af_fetch()] with an advancing `offset`, stopping when a
#' chunk comes back empty or `total_limit` is reached, then row-binds the chunks
#' into one tibble. Columns listed in the API profile's `drop_cols` are removed.
#'
#' @inheritParams af_fetch
#' @param total_limit Maximum number of records to retrieve in total. Default
#'   `Inf` (all available).
#' @param chunk_size Records to request per chunk. Default `50000`.
#' @return A tibble with all retrieved records.
#' @examples
#' \dontrun{
#' api <- af_api("https://www.bigdata.pe.gov.br/api/buscar",
#'               service = "BigDataPE", auth = af_auth_raw(),
#'               pagination = af_paginate_offset("header"),
#'               drop_cols = "Mensagem")
#' af_fetch_all(api, "dengue", total_limit = 500, chunk_size = 100)
#' }
#' @export
af_fetch_all <- function(api, name, total_limit = Inf, chunk_size = 50000L,
                         query = list(), verbosity = 0L) {
  if (!inherits(api, "apifetch_api")) {
    cli::cli_abort("{.arg api} must be an {.cls apifetch_api} object (see {.fn af_api}).")
  }
  if (!is.list(query)) {
    cli::cli_abort("{.arg query} must be a {.cls list}.")
  }

  total_limit <- .safe_as_integer(total_limit, "total_limit")
  chunk_size  <- .safe_as_integer(chunk_size, "chunk_size")
  verbosity   <- .safe_as_integer(verbosity, "verbosity")
  if (!is.infinite(chunk_size) && chunk_size <= 0) {
    cli::cli_abort("{.arg chunk_size} must be a positive whole number.")
  }

  offset <- 0L
  total_fetched <- 0L
  all_data <- list()

  repeat {
    current_limit <- as.integer(min(chunk_size, total_limit - total_fetched))
    if (current_limit <= 0) break

    chunk <- af_fetch(
      api = api,
      name = name,
      limit = current_limit,
      offset = offset,
      query = query,
      verbosity = verbosity
    )

    drop <- intersect(api$drop_cols, names(chunk))
    if (length(drop)) chunk <- dplyr::select(chunk, -dplyr::all_of(drop))

    if (nrow(chunk) == 0L) break

    all_data <- append(all_data, list(chunk))
    total_fetched <- as.integer(total_fetched) + nrow(chunk)

    if (verbosity > 0L) {
      cli::cli_alert_info("Fetched {nrow(chunk)} records (total: {total_fetched}).")
    }

    offset <- as.integer(offset) + nrow(chunk)
    if (total_fetched >= total_limit) break
  }

  combined <- dplyr::bind_rows(all_data)

  if (verbosity > 0L) {
    cli::cli_alert_success("Fetching complete: {nrow(combined)} records retrieved.")
  }

  combined
}

---
title: "Use case: the Big Data PE API"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Use case: the Big Data PE API}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>", eval = FALSE)
```

`apifetch` was extracted from the
[BigDataPE](https://github.com/StrategicProjects/BigDataPE) package, which wraps the
**Big Data PE** platform — a public-data REST API run by the Government of the
State of Pernambuco, Brazil. This vignette shows how that service maps onto a
single `apifetch` profile, and is the canonical worked example of configuring
the package for a real API.

## What is specific about Big Data PE

The Big Data PE API has two conventions that differ from the `apifetch`
defaults:

- **Authentication** sends the token *verbatim* in the `Authorization` header
  (no `Bearer ` prefix) — use `af_auth_raw()`.
- **Pagination** sends `limit`/`offset` as **HTTP headers**, not query
  parameters — use `af_paginate_offset("header")`.

It also returns a status column named `"Mensagem"` that we drop when combining
chunks (`drop_cols = "Mensagem"`), and it is only reachable from the
**PE Conectado** network or VPN, which we surface via `connect_hint`.

## Defining the profile

```{r}
library(apifetch)

bigdatape <- af_api(
  endpoint     = "https://www.bigdata.pe.gov.br/api/buscar",
  service      = "BigDataPE",
  auth         = af_auth_raw(),
  pagination   = af_paginate_offset("header"),
  drop_cols    = "Mensagem",
  connect_hint = "Ensure you are on the PE Conectado network or VPN."
)
```

## Storing a token and fetching

```{r}
af_store_token("dengue", "your-token-here", service = "BigDataPE")

# A single page of 50 records
dengue <- af_fetch(bigdatape, "dengue", limit = 50)

# Everything, in chunks, with a progress message per chunk
dengue_all <- af_fetch_all(
  bigdatape, "dengue",
  chunk_size = 50000,
  verbosity = 1
)
```

## Note on language

Function and parameter names are English (R convention), but the API's response
column names and some values are Portuguese (e.g. `nu_notificacao`,
`"BOA VIAGEM"`). That is intentional — they come straight from the upstream
service.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fetch.R
\name{af_fetch_all}
\alias{af_fetch_all}
\title{Fetch all data from an API in chunks}
\usage{
af_fetch_all(
  api,
  name,
  total_limit = Inf,
  chunk_size = 50000L,
  query = list(),
  verbosity = 0L
)
}
\arguments{
\item{api}{An \code{apifetch_api} object (see \code{\link[=af_api]{af_api()}}).}

\item{name}{The token name to authenticate with (looked up via the API's
\code{service}).}

\item{total_limit}{Maximum number of records to retrieve in total. Default
\code{Inf} (all available).}

\item{chunk_size}{Records to request per chunk. Default \code{50000}.}

\item{query}{A named list of additional query-string filters. Default empty.}

\item{verbosity}{\code{0} (silent, default), \code{1} (progress messages), or \code{2}
(progress plus full HTTP request/response details).}
}
\value{
A tibble with all retrieved records.
}
\description{
Iteratively calls \code{\link[=af_fetch]{af_fetch()}} with an advancing \code{offset}, stopping when a
chunk comes back empty or \code{total_limit} is reached, then row-binds the chunks
into one tibble. Columns listed in the API profile's \code{drop_cols} are removed.
}
\examples{
\dontrun{
api <- af_api("https://www.bigdata.pe.gov.br/api/buscar",
              service = "BigDataPE", auth = af_auth_raw(),
              pagination = af_paginate_offset("header"),
              drop_cols = "Mensagem")
af_fetch_all(api, "dengue", total_limit = 500, chunk_size = 100)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api.R
\name{af_api}
\alias{af_api}
\title{Describe an API endpoint}
\usage{
af_api(
  endpoint,
  service = "apifetch",
  auth = af_auth_bearer(),
  pagination = af_paginate_offset(),
  drop_cols = character(0),
  connect_hint = NULL
)
}
\arguments{
\item{endpoint}{The base API URL.}

\item{service}{Namespace used to look up the token (see \code{\link[=af_get_token]{af_get_token()}}).}

\item{auth}{An \code{apifetch_auth} strategy (see \link{af_auth}). Default
\code{\link[=af_auth_bearer]{af_auth_bearer()}}.}

\item{pagination}{An \code{apifetch_pagination} strategy (see \link{af_paginate}).
Default \code{\link[=af_paginate_offset]{af_paginate_offset()}}.}

\item{drop_cols}{Character vector of response columns to drop after parsing
(e.g. a status column). Default none.}

\item{connect_hint}{Optional extra line shown when a connection error occurs
(e.g. a VPN requirement).}
}
\value{
An \code{apifetch_api} object.
}
\description{
Bundles an endpoint URL with its authentication and pagination strategies and
a namespace \code{service} (used to look up tokens). The resulting object is passed
to \code{\link[=af_fetch]{af_fetch()}} and \code{\link[=af_fetch_all]{af_fetch_all()}}.
}
\examples{
af_api(
  endpoint = "https://www.bigdata.pe.gov.br/api/buscar",
  service = "BigDataPE",
  auth = af_auth_raw(),
  pagination = af_paginate_offset("header"),
  drop_cols = "Mensagem",
  connect_hint = "Ensure you are on the PE Conectado network or VPN."
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{parse_queries}
\alias{parse_queries}
\title{Build a URL with query parameters}
\usage{
parse_queries(url, query_list)
}
\arguments{
\item{url}{The base URL.}

\item{query_list}{A named list of query parameters.}
}
\value{
The URL with the query string appended (or the base URL unchanged
when there are no parameters to add).
}
\description{
Appends a named list of query parameters to a base URL, URL-encoding both
names and values and dropping parameters whose value is the empty string.
}
\examples{
parse_queries("https://example.com", list(a = "1", b = "2"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api.R
\name{af_paginate}
\alias{af_paginate}
\alias{af_paginate_offset}
\alias{af_paginate_none}
\title{Pagination strategies}
\usage{
af_paginate_offset(
  where = c("header", "query"),
  limit_param = "limit",
  offset_param = "offset"
)

af_paginate_none()
}
\arguments{
\item{where}{Either \code{"header"} or \code{"query"}.}

\item{limit_param, offset_param}{Parameter names to use.}
}
\value{
An \code{apifetch_pagination} object.
}
\description{
Constructors describing how \code{limit}/\code{offset} are sent with a request. Pass
the result to the \code{pagination} argument of \code{\link[=af_api]{af_api()}}.
}
\details{
\itemize{
\item \code{af_paginate_offset()}: send \code{limit}/\code{offset} either as HTTP headers
(default, as the Big Data PE API expects) or as URL query parameters.
\item \code{af_paginate_none()}: send no pagination parameters.
}

Non-positive or infinite values are omitted from the request.
}
\examples{
af_paginate_offset("header")
af_paginate_offset("query", limit_param = "per_page")
af_paginate_none()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tokens.R
\name{af_remove_token}
\alias{af_remove_token}
\title{Remove a stored API token}
\usage{
af_remove_token(name, service = "apifetch")
}
\arguments{
\item{name}{The identifier for this token (e.g. a dataset or resource name).}

\item{service}{A namespace prefix grouping tokens for one API. Default
\code{"apifetch"}.}
}
\value{
Invisibly \code{NULL}; called for its side effect.
}
\description{
Removes the token stored for \code{name} under \code{service}. Does nothing (beyond a
warning) when no token is found.
}
\examples{
af_remove_token("dengue", service = "BigDataPE")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api.R
\name{af_auth}
\alias{af_auth}
\alias{af_auth_raw}
\alias{af_auth_bearer}
\alias{af_auth_header}
\alias{af_auth_query}
\title{Authentication strategies}
\usage{
af_auth_raw(header = "Authorization")

af_auth_bearer(header = "Authorization", prefix = "Bearer ")

af_auth_header(header = "X-API-Key")

af_auth_query(param = "api_key")
}
\arguments{
\item{header}{Header name to use.}

\item{prefix}{String prepended to the token (bearer scheme).}

\item{param}{Query-parameter name (query scheme).}
}
\value{
An \code{apifetch_auth} object.
}
\description{
Constructors describing how a token is attached to a request. Pass the result
to the \code{auth} argument of \code{\link[=af_api]{af_api()}}.
}
\details{
\itemize{
\item \code{af_auth_raw()}: send the token verbatim in a header (default
\code{Authorization}). This is what the Big Data PE API expects.
\item \code{af_auth_bearer()}: send \code{"Bearer <token>"} in the \code{Authorization} header.
\item \code{af_auth_header()}: send the token in an arbitrary header (e.g.
\code{X-API-Key}).
\item \code{af_auth_query()}: send the token as a URL query parameter.
}
}
\examples{
af_auth_raw()
af_auth_bearer()
af_auth_header("X-API-Key")
af_auth_query("api_key")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/apifetch-package.R
\docType{package}
\name{apifetch-package}
\alias{apifetch}
\alias{apifetch-package}
\title{apifetch: Token-Authenticated REST API Retrieval Toolkit}
\description{
\if{html}{\figure{logo.svg}{options: style='float: right' alt='logo' width='120'}}

A small, dependency-light toolkit for talking to token-authenticated REST APIs. It manages authentication tokens in process environment variables (never written to disk), builds requests with configurable authentication and pagination strategies, and retrieves paginated data either one page at a time or in chunks combined into a single tibble. The design is API-agnostic: a single 'apifetch_api' profile describes an endpoint together with how it authenticates and paginates, so the same verbs work across different services.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/StrategicProjects/apifetch}
  \item \url{https://strategicprojects.github.io/apifetch/}
  \item Report bugs at \url{https://github.com/StrategicProjects/apifetch/issues}
}

}
\author{
\strong{Maintainer}: André Leite \email{leite@castlab.org}

Authors:
\itemize{
  \item André Leite \email{leite@castlab.org}
  \item Hugo Vasconcelos \email{hugo.vasconcelos@ufpe.br}
  \item Diogo Bezerra \email{diogo.bezerra@ufpe.br}
  \item Marcos Wasilew \email{marcos.wasilew@gmail.com}
  \item Carlos Amorin \email{carlos.agaf@ufpe.br}
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tokens.R
\name{af_store_token}
\alias{af_store_token}
\title{Store an API token in an environment variable}
\usage{
af_store_token(name, token, service = "apifetch")
}
\arguments{
\item{name}{The identifier for this token (e.g. a dataset or resource name).}

\item{token}{The authentication token (character).}

\item{service}{A namespace prefix grouping tokens for one API. Default
\code{"apifetch"}.}
}
\value{
Invisibly \code{NULL}; called for its side effect.
}
\description{
Stores an authentication token in a process environment variable named
\code{"<service>_<name>"}. The token is never written to disk. If a non-empty
variable with that name already exists, the function refuses to overwrite it.
}
\examples{
bdpe <- af_store_token("dengue", "your-token-here", service = "BigDataPE")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fetch.R
\name{af_fetch}
\alias{af_fetch}
\title{Fetch a single page from an API}
\usage{
af_fetch(api, name, limit = Inf, offset = 0L, query = list(), verbosity = 0L)
}
\arguments{
\item{api}{An \code{apifetch_api} object (see \code{\link[=af_api]{af_api()}}).}

\item{name}{The token name to authenticate with (looked up via the API's
\code{service}).}

\item{limit}{Maximum number of records to request. Default \code{Inf} (no limit).
Non-positive or infinite values omit the parameter.}

\item{offset}{Starting record. Default \code{0} (omitted).}

\item{query}{A named list of additional query-string filters. Default empty.}

\item{verbosity}{\code{0} (silent, default), \code{1} (progress messages), or \code{2}
(progress plus full HTTP request/response details).}
}
\value{
A tibble with the parsed response.
}
\description{
Performs one authenticated request against an \code{\link[=af_api]{af_api()}} profile, applying its
pagination strategy, and returns the parsed body as a tibble. HTTP errors and
connection failures are translated into friendly \code{cli} messages.
}
\examples{
\dontrun{
api <- af_api("https://www.bigdata.pe.gov.br/api/buscar",
              service = "BigDataPE", auth = af_auth_raw(),
              pagination = af_paginate_offset("header"))
af_store_token("dengue", "token", service = "BigDataPE")
af_fetch(api, "dengue", limit = 50)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tokens.R
\name{af_get_token}
\alias{af_get_token}
\title{Retrieve a stored API token}
\usage{
af_get_token(name, service = "apifetch")
}
\arguments{
\item{name}{The identifier for this token (e.g. a dataset or resource name).}

\item{service}{A namespace prefix grouping tokens for one API. Default
\code{"apifetch"}.}
}
\value{
The token string, or \code{NULL} if not found.
}
\description{
Retrieves the token stored for \code{name} under \code{service}. Returns \code{NULL}
(with a warning) rather than erroring when no token is found.
}
\examples{
token <- af_get_token("dengue", service = "BigDataPE")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tokens.R
\name{af_list_tokens}
\alias{af_list_tokens}
\title{List stored API tokens}
\usage{
af_list_tokens(service = "apifetch")
}
\arguments{
\item{service}{A namespace prefix grouping tokens for one API. Default
\code{"apifetch"}.}
}
\value{
A character vector of token names, empty if none are found.
}
\description{
Returns the names (without the \code{service} prefix) of all tokens stored for a
given \code{service} in environment variables.
}
\examples{
af_list_tokens(service = "BigDataPE")
}
